import java.io.*;
import java.util.regex.*;
import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.InputSource;

public class PCGentoWerks {

  private File inputfile;
  private File outputfile;
  private String filecontents;
  private NodeList spellexpressions;
  private short spellexpressionslength;
  private Document char_sheet;

  private boolean placeholders;
  private boolean insert_spellexpressions;
  private boolean spellsounds;
  private boolean nobooks;
  private boolean nosaves;
  private boolean nospelllevelnames;

  private List soundlist;
  private File werksdir;

  private boolean verbose;

  private boolean npc;

  Pattern p;
  Matcher m;

  private final String[] spellclasses = {"bard", "cleric", "druid", "paladin",
                                         "ranger", "sorcerer", "wizard"};
  private final byte spellclassescount = (byte) spellclasses.length;
  private final String[] spellclassplaceholders =
      {"$BAL", "$CL", "$DL", "($PL)/2", "($RL)/2", "$SORL", "$WL"};
  private String casterclassplaceholder = "$L";

  public PCGentoWerks(String[] args) {

    verbose = false;

    boolean force = false;
    boolean delete = false;

    placeholders = false;
    insert_spellexpressions = false;
    spellsounds = false;

    nobooks = false;
    nosaves = false;
    nospelllevelnames = false;

    npc = false;

    File spellexpressionfile = new File("PCGentoWerks_spells.xml");
    if (spellexpressionfile.exists() &&
        spellexpressionfile.canRead()) insert_spellexpressions = true;

    if (args.length == 0) {
      System.out.println("ERROR: No arguments specified.\n");
      printhelp();
      System.exit(0);
    }

    // Find last argument and check if it is a valid file

    String lastarg = args[args.length - 1];

    if (lastarg.equals("-v")) {
      printversion();
      System.exit(0);
    }

    inputfile = new File(lastarg);
    if (!inputfile.exists()) {
      System.out.println("ERROR: File not found - "
                         + inputfile.getAbsolutePath() +"\n");
      System.exit(0);
    }
    outputfile = inputfile;

    // READ ARGUMENTS

    if (args.length > 1) {
      for (int i = 0; i < args.length - 1; i++) {
//        System.out.println("\"" + args[i] + "\"");
        if ((args[i].equals("-o")) && (i <= args.length - 3)) {
          File outputdir = new File(args[++i]);
          if (!outputdir.isDirectory()) {
            System.out.println("ERROR: Output directory is not valid.");
            System.exit(0);
          }
          outputfile =
              new File(outputdir + File.separator + inputfile.getName());
          System.out.println("Output set to " + outputfile.toString());
          continue;
        }
        else if (args[i].equals("-d")) {
          delete = true;
          continue;
        }
        else if (args[i].equals("-f")) {
          force = true;
          continue;
        }
        else if (args[i].equals("-p")) {
          placeholders = true;
          continue;
        }
        else if (args[i].equals("-s")) {
          werksdir = new File(args[++i]);
          if (werksdir.isDirectory() && werksdir.canRead())
            spellsounds = true;
          else
            System.out.println(werksdir.toString() +
                               " is not a readable directory.");
          continue;
        }
        else if (args[i].equals("-nobooks")) {
          nobooks = true;
          continue;
        }
        else if (args[i].equals("-nosaves")) {
          nosaves = true;
          continue;
        }
        else if (args[i].equals("-nospelllevelnames")) {
          nospelllevelnames = true;
          continue;
        }
        else if (args[i].equals("-v")) {
          printversion();
          continue;
        }
      }
    }

    if (delete && inputfile.equals(outputfile)) {
      System.out.println("-d ignored. " +
                         "Outputdir must be different from inputdir.");
      delete = false;
    }

    // READ INPUTFILE

    if (verbose) System.out.println("Reading " + inputfile.getName());

    filecontents = readfile(inputfile);

    // See if inputfile is one we like

    if ((filecontents.indexOf("kLoOge.Werks") == -1) && (!force)) {
      System.out.println("Inputfile not created with kLoOge.Werks template.");
      System.exit(0);
    }

    // Check if character is an NPC

    if (filecontents.matches("(?is).*<player>NPC</player>.*")) npc = true;

    // READ SPELL EXPRESSIONS FILE

    if (insert_spellexpressions) {
      if (verbose) System.out.println("Reading PCGentoWerks_spells.xml");
      spellexpressions =
          readXML(readfile(spellexpressionfile)).getElementsByTagName("spell");
      if (spellexpressions == null) {
        System.out.println("Error in PCGentoWerks_spells.xml");
        System.exit(0);
      }
      spellexpressionslength = (short) spellexpressions.getLength();
    }

    // PROCESS INPUTFILE

    if (verbose) System.out.println("Performing initial cleanup...");

    // Replace HTML entities
    replace("&#39;", "'", true);
    replace("&quot;", "\"", true);

    replace("&#230;", "", true);
    replace("&#248;", "", true);
    replace("&#229;", "", true);
    replace("&#198;", "", true);
    replace("&#216;", "", true);
    replace("&#197;", "", true);

    // Convert <type> field contents to lowercase
    p = Pattern.compile("<type>(.*?)</type>");
    m = p.matcher(filecontents);
    if (m.find()) {
      String type = m.group(1);
      if ((type != null) && (type.length() > 2)) {
        String typelower =
            type.substring(0,1) + type.substring(1).toLowerCase();
        replace(escape_regex(type), escape_regex(typelower), false);
      }
    }

    // Remove trailing/leading slashes and zeros in attribute fields
    replace("<age-sex>0?/", "<age-sex>", false);
    replace("/</age-sex>", "</age-sex>", false);
    replace("<height-weight>(0' 0\")?/", "<height-weight>", false);
    replace("/?(0\\s+lbs\\.?)?</height-weight>", "</height-weight>", false);
    replace("<hair-eyes>/", "<hair-eyes>", false);
    replace("/</hair-eyes>", "</hair-eyes>", false);
    replace("<experience>0", "<experience>", false);
    replace("<next-level>0", "<next-level>", false);

    // Remove (1) after item names
    replace(" \\(1\\)</(equip|mi)-name>", "</$1-name>", true);

    // Remove empty locations in item descriptions
    replace("Location: \\. ", "", true);

    // Remove empty skill, equipment and weapon listings
    // (they get output if the character hasn't got
    // any skills, or any equipped or carried weapons)
    replace("(?s)<skill>.{0,10}?<skill-name></skill-name>.*?</skill>",
            "", true);
    replace("(?s)<weapon>.{0,10}?<w-name></w-name>.*?</weapon>",
            "", true);
    replace("(?s)<item>.{0,10}?<equip-name> \\(\\)</equip-name>.*?</item>",
            "", true);
    replace("(?s)<magic-item>.{0,10}?<mi-name> \\(\\)</mi-name>.*?" +
            "</magic-item>", "", true);

    // Remove spot and listen skills if they're
    // listed elsewhere than first in the skills list
    replace("(?s)(<skill-name>Spot</skill-name>.*)" +
            "(<skill>.*?<skill-name>Spot</skill.name>.*?</skill>)",
            "$1", false);
    replace("(?s)(<skill-name>Listen</skill-name>.*)" +
            "(<skill>.*?<skill-name>Listen</skill.name>.*?</skill>)",
            "$1", false);

    // Remove .0 decimal in skill ranks
    replace("\\.0</rank>", "</rank>", true);

    // Convert damage reduction to a plain number and insert the full
    // damage reduction string as a racial trait
    p = Pattern.compile("<damage-reduction>(\\d+/.*?)</damage-reduction>");
    m = p.matcher(filecontents);
    if (m.find()) {
      String damagereduction = m.group(1);
      replace("<racial-traits>", escape_regex("<racial-traits>\r\n"
              + "    <racial-trait>\r\n      <trait-name>"
              + "Damage Reduction: " + damagereduction
              + "</trait-name>\r\n    </racial-trait>\r\n"), false);
      replace("<damage-reduction>(\\d+)/.*?</damage-reduction>",
              "<damage-reduction>$1</damage-reduction>", false);
    }

    // Remove modifier to spell resistance (because PCGen has a fit
    // if you try |SR.NOZERO.INTVAL| on someone without spell resistance,
    // so you can only use |SR.NOZERO| which for some strange reason always
    // adds a modifier before the spell resistance value)
    replace("<spell-resistance>\\+", "<spell-resistance>", false);

    // Convert spells with "At Will" in num-memorized
    replace("(?is)<num-memorized>At Will</num-memorized>.*?<cost>1</cost>",
            "<num-memorized>99</num-memorized>\r\n      <cost>0</cost>",
            true);

    // Remove multiple empty lines
    // (though it doesn't matter to Werks; I'm just anal)
    replace("(\\r?\\n)*\\r?\\n", "\n", true);

    // INSERT PLACEHOLDERS

    if (placeholders) insert_placeholders();

    // Remove DC from spells with no save
    replace("<save>None.*?</save>", "<save>None</save>", true);

    // ADD SNEAK ATTACK AS WEAPON

    p = Pattern.compile("<trait-name>Sneak Attack \\+(\\dd\\d)</trait-name>");
    m = p.matcher(filecontents);
    if (m.find()) {
      String sneakattackdamage = m.group(1);
      replace("(?is)<weapons>(.*?)</weapons>",
              "<weapons>$1\r\n" +
              "    <weapon>\r\n" +
              "      <w-name>Sneak Attack</w-name>\r\n" +
              "      <damage>" + sneakattackdamage + "</damage>\r\n" +
              "    </weapon>\r\n" +
              "  </weapons>\r\n", false);
    }

    // ANONYMISE WEAPON NAMES

    if (npc) anonymise_weapons();

    // CREATE XML DOM DOCUMENT

    if (verbose) System.out.println("Constructing XML document...");

    char_sheet = readXML(filecontents);

    // PROCESS SPELLS

    if (verbose) System.out.println("Processing spells...");

    // Shorten class names in spell levels (e.g. "Cleric 1" becomes "Clr 1")
/*    replace("(?is)<(cast-)?level>Bard (\\d+)</(cast-)?level>",
            "<$1level>Brd $2</$3level>", true);
    replace("(?is)<(cast-)?level>Cleric (\\d+)</(cast-)?level>",
            "<$1level>Clr $2</$3level>", true);
    replace("(?is)<(cast-)?level>Druid (\\d+)</(cast-)?level>",
            "<$1level>Drd $2</$3level>", true);
    replace("(?is)<(cast-)?level>Paladin (\\d+)</(cast-)?level>",
            "<$1level>Pal $2</$3level>", true);
    replace("(?is)<(cast-)?level>Ranger (\\d+)</(cast-)?level>",
            "<$1level>Rgr $2</$3level>", true);
    replace("(?is)<(cast-)?level>Sorcerer (\\d+)</(cast-)?level>",
            "<$1level>Sor $2</$3level>", true);
    replace("(?is)<(cast-)?level>Wizard (\\d+)</(cast-)?level>",
            "<$1level>Wiz $2</$3level>", true);
*/
    process_spells();

    // CREATE SPELLBOOKS

    if (!nobooks)
      create_spellbooks();

    // PERFORM VARIOUS CLEANUP

    if (verbose)
      System.out.println("Cleaning data...");

    filecontents = filecontents.substring(0,
      filecontents.indexOf("<char-sheet-3rd-ed>"))
      + char_sheet.getDocumentElement().toString();

    // Remove class names in spell levels if requested
    if (nospelllevelnames) {
      replace("<(cast-)?level>.*?(\\d+)</(cast-)?level>",
              "<$1level>$2</$3level>", true);
    }

    // Remove spellbooks - Werks can't use them, so they just take up space
    replace("(?is)<spellbook>.*?</spellbook>", "", true);

    // WRITE FILE

    if (verbose)
      System.out.println("Writing file...");

    try {
        BufferedWriter out = new BufferedWriter(new FileWriter(outputfile));
        out.write(filecontents);
        out.close();
    }
    catch (IOException e) {
      System.out.println("Unable to write file. " +
                         "Something is very wrong!\nRun, run for your lives!!");
    }

    if (delete) inputfile.delete();

    System.out.println("Done.");
  }

  public static void main(String[] args) {
    PCGentoWerks app = new PCGentoWerks(args);
  }

  private void process_spells() {

    NodeList spells = char_sheet.getElementsByTagName("spell");
    short spellcount = (short) spells.getLength();

    // If the character hasn't got any spells, then don't bother...
    if (spellcount < 1) return;

    // Read sounds
    if (spellsounds) {
      File sounddir = new File(werksdir.toString() + File.separator + "sounds");
      if (!sounddir.isDirectory() || !sounddir.canRead()) {
        System.out.println("Can't read directory " + sounddir.toString());
        spellsounds = false;
      }
      else {
        soundlist = get_files(sounddir, ".ogg");
        soundlist.addAll(get_files(sounddir, ".mp3"));
        soundlist.addAll(get_files(sounddir, ".wav"));
      }
    }

    // Create new nodes to hold the processed spells and custom expressions
    Node newSpells = char_sheet.createElement("spells");
    Node customExpressions = char_sheet.createElement("custom-exprs");

    Node soundNode = char_sheet.createElement("sound");
    Node soundNodeText = char_sheet.createTextNode(null);
    soundNode.appendChild(soundNodeText);

    // Process each of the character's spells
    for (short i = 0; i < spellcount; i++) {

      Node spell = spells.item(i);
      Node oldspell = spell;

      soundNode.getFirstChild().setNodeValue(null);

      String spellname = get_spellname(spell);

      // If the -s switch was used, add matching spell sound, if any
      if (spellsounds) {
        String spellsound = get_sound(spellname);
        if (spellsound != null)
          soundNode.getFirstChild().setNodeValue(spellsound);
      }

      // Find the spellcasting class for the spell
      String casterclass =
          get_tagContents("level", spell).replaceFirst(" \\d+", "");

      for (byte j = 0; j < spellclassescount; j++) {
        if (spellclasses[j].equalsIgnoreCase(casterclass)) {
          casterclassplaceholder = spellclassplaceholders[j];
          break;
        }
      }

      // If no known spellcasting class placeholder
      // is found for this spell, use $L
      if (casterclassplaceholder.length() == 0) casterclassplaceholder = "$L";

      // Find and process the expression for this spell
      Node expressionNode = get_childNode("expr", spell);
      String spellexpression = expressionNode.getFirstChild().toString();
      Node spelldata = null;
      String newExpression = null;

      // Look up this spellname in PCGentoWerks_spells.xml (if it exists)
      if (insert_spellexpressions) {
        spelldata = get_spelldata(spellname);
        if (spelldata != null) {
          newExpression = get_spellexpression(spelldata);

          // Add any <custom-expr> to the custom expressions node
          add_customExpressions(spelldata, customExpressions);

          // Add any effects or sounds for the current spell
          Node effect = get_childNode("effect", spelldata);
          if (effect != null)
            spell.appendChild(char_sheet.importNode(effect, true));
          Node sound = get_childNode("sound", spelldata);
          if (sound != null)
            soundNode = char_sheet.importNode(sound, true);
        }
      }

      // If the spellfile contains an expression for the spell
      if (newExpression != null) {
        spellexpression = newExpression;

        // Replace $L in expression with casterlevelplaceholder
        spellexpression = replace_spellcasterplaceholder(spellexpression);
      }
      else {
        // Process the spell description from PCGen
        spellexpression = insert_expressions(spellexpression);
        if (placeholders)
          spellexpression = insert_spellcasterplaceholder(spellexpression);
      }

      // If a sound was found for this spell, add it to the spell node
      if (soundNode.getFirstChild().getNodeValue().length() != 0)
        spell.appendChild(soundNode);

      // Apprend spell saving throw information to expression
      if (!nosaves) {
        String savingthrow = get_tagContents("save", spell);
        if (!spellexpression.endsWith(".")) spellexpression += ".";
        spellexpression += " Save: " + savingthrow;
      }

      // Delete save node
      spell.removeChild(get_childNode("save", spell));

      // Create new spellnode with new expression
      expressionNode.getFirstChild().setNodeValue(spellexpression);

      Node newSpellNode = spell.cloneNode(true);
      newSpellNode.replaceChild(expressionNode,
                                get_childNode("expr", newSpellNode));

      newSpells.appendChild(newSpellNode);
    }

    // Add new spells node and custom expressions node
    // (if not empty) to char_sheet

    char_sheet.getDocumentElement().replaceChild(newSpells,
      get_childNode("spells", char_sheet.getDocumentElement()));

    if (customExpressions.hasChildNodes())
      char_sheet.getDocumentElement().appendChild(customExpressions);
  }

  private String insert_expressions(String spelldescription) {

    return spelldescription.replaceAll("(\\d+d\\d+)([+-]\\d+)?", "{$1$2}");
  }

  private void insert_placeholders() {

    if (verbose)
      System.out.println("Inserting placeholders in weapons and skills...");

    final short[] AbMod = new short[6];

    final byte STR = 0;
    final byte DEX = 1;
    final byte CON = 2;
    final byte INT = 3;
    final byte WIS = 4;
    final byte CHA = 5;

    // Ability modifiers

    final String[] abilities = {"STR", "DEX", "CON", "INT", "WIS", "CHA"};
    final String[] placeholders = {"$S", "$D", "$C", "$I", "$W", "$CH"};

    // Find ability modifiers

    String modifier = "0";

    for (byte i = 0; i < abilities.length; i++) {
      p = Pattern.compile("(?i)<" + abilities[i] + "-mod>\\+?(-?\\d+)</");
      m = p.matcher(filecontents);
      if (m.find()) {
        modifier = m.group(1);
      }
      else {
        modifier = "0";
      }
      AbMod[i] = Byte.parseByte(modifier);
    }

    // Find weapon finesse feats

    String[] finesseweapons = {};

    p = Pattern.compile("(?si)<feat-name>Weapon Finesse "+
                        "\\((.*?)\\)</feat-name>");
    m = p.matcher(filecontents);

    if (m.find()) {
      String weapons = m.group(1);
      finesseweapons = weapons.split(", ");
    }

    // Find each weapon

    p = Pattern.compile("(?si)<weapon>(.*?)</weapon>");
    m = p.matcher(filecontents);

    while (m.find()) {

      String weapondata = m.group(1);
      String new_weapondata = weapondata;

      String weaponname = "";
      String attacks = "";
      String damage = "";

      boolean ranged = false;
      boolean thrown = false;
      boolean finessed = false;
      boolean mightybow = false;
      boolean xbow = false;
      boolean twohand = false;
      boolean offhand = false;

      byte mightybow_maxstr = -1;

      // Read weapon name, attack values, and damage values

      Pattern p_wp = Pattern.compile("(?si)<w-name>(.+?)</w-name>.*?" +
                                     "<atk-mod>(.+?)</atk-mod>.*?" +
                                     "<damage>(.*?)</damage>");
      Matcher m_wp = p_wp.matcher(weapondata);

      if (m_wp.find()) {

        weaponname = m_wp.group(1);
        attacks = m_wp.group(2);
        damage = m_wp.group(3);

        // Check if weapon is ranged and thrown

        if ((weapondata.matches("(?is).*<range>.*")) ||
            (weaponname.matches("(?is).*Ranged.*"))) ranged = true;
        if (weaponname.matches("(?is).*\\(Thrown\\).*")) thrown = true;

        // Check if it is a mighty composite bow

        if ((ranged) && (is_mightybow(weaponname))) {
          mightybow = true;
          mightybow_maxstr = get_mightybow_maxstr(weaponname);
        }

        // Check if it is a crowssbow

        if (weaponname.matches("(?is).*Crossbow.*"))
          xbow = true;

        // Check if there is a weapon finesse feat for this weapon name

        for (byte i = 0; i < finesseweapons.length; i++) {
          if (weaponname.matches("(?i)" + finesseweapons[i] + ".*")) finessed = true;
        }

        // Check if weapon is equipped in both hands or offhand

        if (weapondata.matches("(?is).*<hand>Both.*")) twohand = true;
        if (weapondata.matches("(?is).*<hand>Off-hand.*")) offhand = true;

        // Insert placeholders in attack values
        // ($S if melee, $D if ranged or finessed)

        String placeholder = placeholders[STR];
        short abmodifier = AbMod[STR];

        if (ranged || finessed) {
          placeholder = placeholders[DEX];
          abmodifier = AbMod[DEX];
        }

        String[] atk_modifiers = attacks.split("/");
        String new_modifiers = "";

        for (byte i = 0; i < atk_modifiers.length; i++) {
          new_modifiers += insert_placeholder(placeholder,
                                              abmodifier, atk_modifiers[i]);
          new_modifiers += "/";
        }

        new_modifiers = new_modifiers.substring(0, new_modifiers.length() - 1);

        new_weapondata = new_weapondata.replaceFirst("<atk-mod>" +
            escape_regex(attacks) + "</atk-mod>",
            "<atk-mod>" + escape_regex(new_modifiers) + "</atk-mod>");

        // Insert placeholders in damage values

        String[] damage_dice = damage.split("/");
        String new_damage = "";

        for (byte i = 0; i < damage_dice.length; i++) {
          Pattern p_dmg = Pattern.compile("\\d+d\\d+([-+]\\d+)");
          Matcher m_dmg = p_dmg.matcher(damage_dice[i]);
          modifier = "";

          if (m_dmg.find()) {
            modifier = m_dmg.group(1);
          }
          else {
            modifier = "+0";
          }

          if (ranged && !thrown) {

            // If mighty composite bow
            if (mightybow && (mightybow_maxstr > 0)) {
              if (AbMod[STR] > mightybow_maxstr) {
                modifier = insert_placeholder(placeholders[STR],
                                              mightybow_maxstr, modifier);
              }
              else {
                modifier = insert_placeholder(placeholders[STR], AbMod[STR],
                                              modifier);
              }
              modifier = modifier.replaceFirst("\\$S", "\\(\\$S:"
                                               + mightybow_maxstr + "\\)");
            }

            // If a str penalty has been subtracted from ranged weapon damage
            else if (AbMod[STR] < 0 && !xbow) {
              modifier = insert_placeholder(placeholders[STR],
                                            AbMod[STR], modifier);
              modifier = modifier.replaceFirst("\\$S", "\\(\\$S:0\\)");
            }

            // If no str modifier has been added to ranged weapon damage
            else {
              if (modifier == "+0") modifier = "";
              if (!xbow)
                modifier += "($S:0)";
            }

          }
          // If weapon is melee or thrown
          else {
            // If weapon is equipped in both hands
            if (twohand) {
              modifier = insert_placeholder(placeholders[STR],
                (short) Math.floor(AbMod[STR] * 1.5f), modifier);
              modifier = modifier.replaceFirst("\\$S",
                                               escape_regex("+(($S)*1.5)"));
            }
            // If weapon is equipped in off-hand
            else if (offhand) {
              modifier = insert_placeholder(placeholders[STR],
                (short) Math.floor(AbMod[STR] * 0.5f), modifier);
              modifier = modifier.replaceFirst("\\$S",
                                               escape_regex("+(($S)*.5)"));
            }
            // If it is just a plain melee or thrown weapon
            else {
              modifier = insert_placeholder(placeholders[STR],
                                            AbMod[STR], modifier);
            }
          }

          new_damage += damage_dice[i].replaceFirst(
              "(\\d+d\\d+).*", "$1" + escape_regex(modifier)) + "/";
        }

        new_damage = new_damage.substring(0, new_damage.length() - 1);

        new_weapondata = new_weapondata.replaceFirst("<damage>"
            + escape_regex(damage) + "</damage>", "<damage>"
            + escape_regex(new_damage) + "</damage>");

        replace(escape_regex(weapondata), escape_regex(new_weapondata), false);
      }
    }

    // Add placeholders to skills

    p = Pattern.compile("(?is)<skill>(.*?)</skill>");
    m = p.matcher(filecontents);

    while (m.find()) {

      String skilldata = m.group(1);
      String new_skilldata = skilldata;

      Pattern p_sk = Pattern.compile("(?is)<total>(.+?)</total>.*?" +
                                     "<ability>(.+?)</ability>");
      Matcher m_sk = p_sk.matcher(skilldata);

      if (m_sk.find()) {

        String skill_modifier = m_sk.group(1);
        String skill_ability = m_sk.group(2);

        for (byte i = 0; i < abilities.length; i++) {

          if (skill_ability.equalsIgnoreCase(abilities[i])) {
            String new_skill_modifier = insert_placeholder(placeholders[i],
                AbMod[i], skill_modifier);
            new_skilldata = new_skilldata.replaceFirst("<total>(.+?)</total>",
                "<total>" + escape_regex(new_skill_modifier) + "</total>");
            replace(escape_regex(skilldata),
                    escape_regex(new_skilldata), false);
            break;
          }
        }
      }
    }

    // Add placeholders to saves

    p = Pattern.compile("(?is)<(ref|fort|will)>([-+0-9]+)<");
    m = p.matcher(filecontents);

    while (m.find()) {

      String saveabilityplaceholder = "";
      short saveabilitymod = 0;

      if (m.group(1).equals("ref")) {
        saveabilityplaceholder = placeholders[DEX];
        saveabilitymod = AbMod[DEX];
      }
      else if (m.group(1).equals("fort")) {
        saveabilityplaceholder = placeholders[CON];
        saveabilitymod = AbMod[CON];
      }
      else if (m.group(1).equals("will")) {
        saveabilityplaceholder = placeholders[WIS];
        saveabilitymod = AbMod[WIS];
      }

      String newSave = insert_placeholder(saveabilityplaceholder,
                                          saveabilitymod, m.group(2));
      replace(escape_regex("<" + m.group(1) + ">" + m.group(2)),
              escape_regex("<" + m.group(1) + ">" + newSave), false);
    }

    // If character has paladin levels and positive cha modifier,
    // add cha placeholder to saves

    if ((filecontents.indexOf("<paladin-level>") > -1) && (AbMod[CHA] > 0)) {

      p = Pattern.compile("(?is)<(ref|fort|will)>([-+0-9]+)(.*?)<");
      m = p.matcher(filecontents);

      while (m.find()) {
        String newSave = insert_placeholder(placeholders[CHA], AbMod[CHA],
                                            m.group(2));
        replace(escape_regex("<" + m.group(1) + ">" + m.group(2) + m.group(3)),
                escape_regex("<" + m.group(1) + ">" + newSave + m.group(3)),
                false);
      }
    }

    // Add placeholders to spell saving throw DCs

    // Abilities affecting spell DC for the classes
    // bard, cleric, druid, paladin, ranger, sorcerer, wizard
    final byte[] spellabilities = {CHA, WIS, WIS, WIS, WIS, CHA, INT};

    p = Pattern.compile("(?is)<spell>(.*?)<level>(.*?)</level>(.*?)" +
                        "<save>(.*?DC )(\\d+)</save>");
    m = p.matcher(filecontents);

    while (m.find()) {
      if (m.group(4).startsWith("None")) continue;
      String saveDC = m.group(5);
      String spellClass = m.group(2).replaceFirst(" \\d+", "");

      byte spellDCability = -1;
      for (byte i = 0; i < spellclassescount; i++) {
        if (spellclasses[i].equalsIgnoreCase(spellClass)) {
          spellDCability = spellabilities[i];
          break;
        }
      }
      if (spellDCability > -1) {
        saveDC = insert_placeholder(placeholders[spellDCability],
                                    AbMod[spellDCability], saveDC);
        if (saveDC.startsWith("+")) saveDC = saveDC.replaceFirst("\\+", "");
        replace("(?is)" + escape_regex(m.group(0)),
                escape_regex("<spell>" + m.group(1) + "<level>" + m.group(2)
                + "</level>" + m.group(3) + "<save>" + m.group(4)
                + "{" + saveDC + "}</save>"), false);
      }
    }

    // Add placeholders to armour class
    p = Pattern.compile("(?is)<ac-final>(\\d+)</ac-final>.*?" +
                        "<touch-ac>(\\d+)</touch-ac>.*?" +
                        "<flat-footed-ac>(\\d+)</flat-footed-ac>");
    m = p.matcher(filecontents);
    m.find();

    String AC = m.group(1);
    String touchAC = m.group(2);
    String flatAC = m.group(3);

    String newAC = AC;
    String newTouchAC = touchAC;
    String newFlatAC = flatAC;

    boolean uncannyDodge = filecontents.matches("(?is).*?<trait-name>" +
      "Uncanny Dodge \\(Dex bonus to AC\\)</trait-name>.*?");

    // Find the max dex bonus to AC (if any)
    short dexmodifier = AbMod[DEX];

    p = Pattern.compile("<ac-maxdex>\\+?(\\d+)</ac-maxdex>");
    m = p.matcher(filecontents);

    if (m.find()) {
      // If there is a max dex value affecting AC
      byte maxdex = Byte.parseByte(m.group(1));
      if (dexmodifier > maxdex) dexmodifier = maxdex;
      newAC = insert_placeholder("$D", dexmodifier, AC) + ":" + maxdex;
      newTouchAC = insert_placeholder("$D", dexmodifier, touchAC)
          + ":" + maxdex;
      if (uncannyDodge) newFlatAC = insert_placeholder("$D", dexmodifier,
            flatAC) + ":" + maxdex;
    }
    else {
      newAC = insert_placeholder("$D", dexmodifier, AC);
      newTouchAC = insert_placeholder("$D", dexmodifier, touchAC);
      if (uncannyDodge) newFlatAC = insert_placeholder("$D", dexmodifier,
            flatAC);
    }

    newAC = newAC.replaceFirst("\\+", "");
    newTouchAC = newTouchAC.replaceFirst("\\+", "");
    newFlatAC = newFlatAC.replaceFirst("\\+", "");

    replace("<ac-final>" + AC + "</ac-final>",
            escape_regex("<ac-final>" + newAC + "</ac-final>"), false);
    replace("<touch-ac>" + touchAC + "</touch-ac>",
            escape_regex("<touch-ac>" + newTouchAC + "</touch-ac>"), false);
    replace("<flat-footed-ac>" + flatAC + "</flat-footed-ac>",
            escape_regex("<flat-footed-ac>" + newFlatAC + "</flat-footed-ac>"),
            false);
  }

  private String insert_spellcasterplaceholder(String expression) {

    // Only parse expressions containing CASTERLEVEL or (something)/level

    boolean hascasterlevel = false;
    boolean hasslashlevel = false;

    if (expression.indexOf("CASTERLEVEL") > -1) hascasterlevel = true;
    if (expression.indexOf("/level") > -1) hasslashlevel = true;

    if (!hascasterlevel && !hasslashlevel) return expression;

    Pattern p_sp;
    Matcher m_sp;

    // Process expressions containing (CASTERLEVEL), inserting
    // the general $L placeholder (which will be replaced later)

    if (hascasterlevel) {

      // Insert $L in place of CASTELEVEL
      expression = expression.replaceAll("CASTERLEVEL", "\\$L");

      // Construct $L:XX expressions where possible
      expression = expression.replaceAll("\\(\\$L\\)MIN(\\d+)", "\\$L:$1");

      // Put parenthesis around $L if it is multiplied or divided
      expression = expression.replaceAll("(\\d+\\*)\\$L", "$1\\(\\$L\\)");
      expression = expression.replaceAll("\\$L([*/]\\d+)", "\\(\\$L\\)$1");

      // Put { and } around expressions
      expression = expression.replaceAll("([ +])([-+*/()0-9d$L:]{3,})([ %,.;])",
                                         "$1\\{$2\\}$3");
      expression = expression.replaceAll("^([-+*/()0-9d$L:]{3,})([ %,.;])",
                                         "\\{$1\\}$2");

      // Remove superflous start and end parenthesis
      expression = expression.replaceAll("\\{\\((.*?)\\)\\}", "\\{$1\\}");

      // Merge adjacent expressions
      expression = expression.replaceAll("\\{\\+", "\\+\\{");
      expression = expression.replaceAll("\\}\\s?\\+\\{", "\\+");

      // Replace $L with specific casterlevelplaceholder
      expression = replace_spellcasterplaceholder(expression);
    }

    // Find (some digits)(some text)/level and insert placeholder

    if (hasslashlevel) {

      p_sp = Pattern.compile("(?i)\\s(\\d+|one)([^0-9]*?)/level\\.?");
      m_sp = p_sp.matcher(expression);

      while (m_sp.find()) {
        String digits = m_sp.group(1);
        String text = m_sp.group(2);
        String new_stuff = "";
        String old_stuff = m_sp.group(0);

        if (digits.equals("1") || digits.equalsIgnoreCase("one"))
          new_stuff = " {" + casterclassplaceholder + "}" + text;
        else
          new_stuff =
              " {" + digits + "*(" + casterclassplaceholder + ")}" + text;

        expression = expression.replaceFirst(escape_regex(old_stuff),
                                             escape_regex(new_stuff));
      }
    }

    return expression;
  }

  private String replace_spellcasterplaceholder(String expression) {

    if (casterclassplaceholder.equals("$L")) return expression;

    // If the level placeholder is ($PL)/2 or similar,
    // it needs "special treatment", because it can't
    // be inserted directly in place of $L in eg. $L:10.
    // Expressions such as $L:10 needs to be converted to
    // ($PL:20)/2 (for example)

    if (casterclassplaceholder.matches("\\(\\$[A-Z]+\\)/\\d+")) {

      // Find the placeholder and division factor
      Pattern p_sp = Pattern.compile("(?i)(\\$[A-Z]+)\\)/(\\d+)");
      Matcher m_sp = p_sp.matcher(casterclassplaceholder);

      m_sp.find();

      String placeholder = m_sp.group(1);
      byte factor = Byte.parseByte(m_sp.group(2));

      // Find any $L:XX placeholders and replace them
      p_sp = Pattern.compile("(?i)\\$L:(\\d+)");
      m_sp = p_sp.matcher(expression);

      while (m_sp.find()) {
        byte maxvalue = Byte.parseByte(m_sp.group(1));
        String old_placeholder = "$L:" + maxvalue;
        String new_placeholder = "(" + placeholder + ":"
                                 + maxvalue*factor + ")/" + factor;
        expression = expression.replaceFirst(escape_regex(old_placeholder),
                                             escape_regex(new_placeholder));
      }

      // Replace any remaining placeholders
      expression = expression.replaceAll(escape_regex("$L"),
                                         escape_regex(casterclassplaceholder));
    }
    else {
      expression = expression.replaceAll(escape_regex("$L"),
                                         escape_regex(casterclassplaceholder));
    }

    return expression;
  }

  private void anonymise_weapons() {
    p = Pattern.compile("<w-name>(.*?)</w-name>");
    m = p.matcher(filecontents);

    while (m.find()) {
      String weaponname = m.group(1);
      String newweaponname = weaponname;
      newweaponname = newweaponname.replaceAll("\\s?[\\+-]\\d", "");
      newweaponname = newweaponname.replaceAll("(?i)\\s?Mighty/?", "");
      newweaponname = newweaponname.replaceAll("(?i)Masterwork", "");

      newweaponname = newweaponname.replaceAll("//", "");
      newweaponname = newweaponname.replaceAll(escape_regex("()"), "");

      replace("<w-name>" + escape_regex(weaponname) + "</w-name>",
              "<w-name>" + escape_regex(newweaponname) + "</w-name>",
              false);
    }
  }

  // Thank you for reading this far in the sourcecode.

  private void create_spellbooks() {

    NodeList spellbooks = char_sheet.getElementsByTagName("spellbook");
    short spellbookcount = (short) spellbooks.getLength();

    if (spellbookcount < 1) return;

    if (verbose) System.out.println("Creating spellbook files...");

    // Get all known spells
    NodeList spells = char_sheet.getElementsByTagName("spell");

    // Process each spellbook
    for (short i = 0; i < spellbookcount; i++) {

      Node spellbook = spellbooks.item(i);

      // Skip empty spellbooks
      if (get_childNode("spellbook-spell", spellbook) == null) continue;

      // Get spellbook attributes
      String spellbookclass = get_tagContents("spellbook-class", spellbook);
      String spellbookname = get_tagContents("spellbook-name", spellbook);
      spellbookname = spellbookname.replaceAll("\\s", "_");

      // Make an empty spellbook for holding all known spells for the class
      Node newSpellBook = spellbook.cloneNode(false);

      // Find all known spells for the spellbook's class and add them to the
      // new spellbook
      short spellcount = (short) spells.getLength();

      for (short j = 0; j < spellcount; j++) {
        Node spell = spells.item(j);
        String spellclass =
            get_tagContents("level", spell).replaceFirst(" \\d+", "");
        if (!spellclass.equalsIgnoreCase(spellbookclass)) continue;
        newSpellBook.appendChild(spell.cloneNode(true));
      }

      // Get the spells memorised in the spellbook, and add the # memorised
      // to the new spellbook
      for (Node stuff = spellbook.getFirstChild(); stuff != null;
           stuff = stuff.getNextSibling()) {

        if (!stuff.getNodeName().equals("spellbook-spell")) continue;

        String spellname = get_spellname(stuff);

        // Make the spellname into a regex pattern
        // (so that "Bless [Widen Spell]" will match "Bless", for example)
        String spellnamepattern = spellname.replaceAll("\\[.*?\\]", "");
        spellnamepattern = escape_regex(spellnamepattern);
        spellnamepattern = spellnamepattern.replaceAll(" ", "\\\\s?");

        // Find the node matching the spellname in the new spellbook
        for (Node spell = newSpellBook.getFirstChild(); spell != null;
             spell = spell.getNextSibling()) {

          if (!get_spellname(spell).matches(spellnamepattern)) continue;

          // Set the spellname (in case it contains metamagic)
          get_childNode("spell-name", spell).setNodeValue(spellname);
          // Add the memorised node
          spell.appendChild(get_childNode("num-memorized",
                                          stuff).cloneNode(true));
        }
      }

      // Right, so now we have a spellbook with all known spells for the class,
      // and number of spells memorised. That has to be converted to a
      // tab-delimited string, which we can later import into Werks.

      Character tab = new Character('\u0009');

      StringBuffer finalSpellBook = new StringBuffer();
      finalSpellBook.append("# Name" + tab + "Level" + tab + "Cast Level" + tab
                            + "Expression" + tab + "# Memorized" + tab + "Cost"
                            + tab + "Effect" + tab + "Sound\r\n");

      for (Node spell = newSpellBook.getFirstChild(); spell != null;
           spell = spell.getNextSibling()) {
        if (!spell.getNodeName().equals("spell")) continue;

        finalSpellBook.append(get_spellname(spell) + tab);

        if (!nospelllevelnames) {
          finalSpellBook.append(get_tagContents("level", spell) + tab);
          finalSpellBook.append(get_tagContents("cast-level", spell) + tab);
        }
        else {
          String spelllevel =
              get_tagContents("level", spell).replaceFirst(".*?(\\d+)", "$1");
          String casterlevel =
              get_tagContents("cast-level",
                              spell).replaceFirst(".*?(\\d+)", "$1");
          finalSpellBook.append(spelllevel + tab);
          finalSpellBook.append(casterlevel + tab);
        }

        finalSpellBook.append(get_tagContents("expr", spell) + tab);
        finalSpellBook.append(get_tagContents("num-memorized", spell) + tab);
        finalSpellBook.append(get_tagContents("cost", spell) + tab);
        finalSpellBook.append(get_tagContents("effect", spell) + tab);
        finalSpellBook.append(get_tagContents("sound", spell));
        finalSpellBook.append("\r\n");
      }
      String finalFinalSpellBook = finalSpellBook.toString();
      finalFinalSpellBook =
          finalFinalSpellBook.replaceAll(tab.toString() + "null",
                                         tab.toString());

      // Write spellbookfile

      String spellbookfilename = outputfile.toString().replaceAll("\\.xml", "")
          + "_" + spellbookclass + "spellbook_" + spellbookname + ".txt";

      File spellbookfile = new File(spellbookfilename);
      try {
          BufferedWriter out =
              new BufferedWriter(new FileWriter(spellbookfile));
          out.write(finalFinalSpellBook);
          out.close();
      }
      catch (IOException e) {
        System.out.println("Unable to write spellbookfile. How odd...");
      }
    }
  }

  private String insert_placeholder(String placeholder,
                                    short modifier, String subject) {

    if (subject.startsWith("+")) subject = subject.substring(1);
    short value = new Short(subject).shortValue();
    if (value == modifier) return placeholder;
    String placeholderstring = "";
    value -= modifier;
    if (value >= 0) placeholderstring += "+";
    placeholderstring += Short.toString(value) + placeholder;
    return placeholderstring;

  }

  private boolean is_mightybow(String weaponname) {

    if ((weaponname.matches("(?is).*Mighty.*"))
        && (weaponname.matches("(?is).*Composite.*"))) {
      return true;
    }
    else {
      return false;
    }
  }

  private byte get_mightybow_maxstr(String weaponname) {

    Pattern p_mb = Pattern.compile("(?i)\\(\\+(\\d)\\s(Str|Mighty)");
    Matcher m_mb = p_mb.matcher(weaponname);
    if (m_mb.find()) return new Byte(m_mb.group(1)).byteValue();
    else return -1;
  }

  private Node get_spelldata(String spellname) {

    for (short i = 0; i < spellexpressionslength; i++) {
      Node spellnode = spellexpressions.item(i);
      if (spellname.equalsIgnoreCase(get_spellname(spellnode)))
          return spellnode;
    }
    return null;
  }

  private Node get_childNode(String nodeName, Node node) {

    NodeList nodeContents = node.getChildNodes();

    for (Node stuff = node.getFirstChild(); stuff != null;
         stuff = stuff.getNextSibling()) {
      if (stuff.getNodeName().equals(nodeName))
        return stuff;
    }
    return null;
  }

  private String get_tagContents(String tag, Node node) {

    Node tagNode = get_childNode(tag, node);
    if (tagNode == null) return null;
    if (!tagNode.hasChildNodes()) return null;
    return tagNode.getFirstChild().getNodeValue();
  }

  private String get_spellname(Node spell) {

    Node namenode = get_childNode("spell-name", spell);
    if (namenode != null)
      return namenode.getFirstChild().toString();
    else
      return null;
  }

  private String get_spellexpression(Node spell) {

    Node spellexpression = get_childNode("expr", spell);
    if (spellexpression != null) {
      return spellexpression.getFirstChild().toString();
    }
    else
      return null;
  }

  // Adds custom expressions to the <custom-exprs> node in the charsheet
  private void add_customExpressions(Node spell, Node custExprNode) {

    for (Node node = spell.getFirstChild(); node != null;
         node = node.getNextSibling()) {
      if (node.getNodeName().equals("custom-expr")) {

        // Replace $L with the casterlevel placeholder
        if (placeholders) {
          String expression =
              get_childNode("expr-expr", node).getFirstChild().toString();
          expression = replace_spellcasterplaceholder(expression);
          get_childNode("expr-expr",
                        node).getFirstChild().setNodeValue(expression);
        }

        Node custExpr = char_sheet.importNode(node, true);
        custExprNode.appendChild(custExpr);

      }
    }
  }

  private List get_files(File startdir, String extension) {
    List result = new ArrayList();
    List FilesAndDirs = Arrays.asList(startdir.listFiles());
    Iterator FileIterator = FilesAndDirs.iterator();
    short werksdirlen = (short) werksdir.toString().length();
    werksdirlen++;

    while (FileIterator.hasNext()) {
      File currentFile = (File) FileIterator.next();
      String fileName = currentFile.getName();

      // If the filename ends with the required extension,
      // store the filename (minus extension) and the relative path.
      short len = (short) fileName.length();
      if (fileName.substring(len - extension.length(),
                             len).equalsIgnoreCase(extension)) {
        fileName = fileName.substring(0, len - extension.length());
        // Replace any underscores or dots with spaces
        fileName = fileName.replaceAll("[_.]", " ");

        // Get the path and filename, and remove the werksdir part,
        // leaving the realtive filepath from the werksdir
        String filePath = currentFile.getPath();
        filePath = filePath.substring(werksdirlen, filePath.length());

        // Add the filename and relative path+filename to the result list
        result.add(new String[] {fileName, filePath});
        continue;
      }

      // Recursively read any subdirectories
      if (currentFile.isDirectory() && currentFile.canRead()) {
        result.addAll(get_files(currentFile, extension));
      }
    }

    return result;
  }

  private String get_sound(String soundname) {

    // Clean up whitespace and asterisks in name

    soundname = soundname.replaceAll("\\*", "");
    soundname = soundname.trim();

    Iterator cousinIt = soundlist.iterator();

    while (cousinIt.hasNext()) {
      String[] sounddata = (String[]) cousinIt.next();
      if (sounddata[0].equalsIgnoreCase(soundname))
        return sounddata[1];
    }
    return null;
  }

  private String escape_regex(String regex) {

    // Damn, that's a lot of backslashes...

    regex = regex.replaceAll("\\.", "\\\\.");
    regex = regex.replaceAll("\\+", "\\\\+");
    regex = regex.replaceAll("\\*", "\\\\*");
    regex = regex.replaceAll("\\?", "\\\\?");
    regex = regex.replaceAll("\\$", "\\\\\\$");
    regex = regex.replaceAll("\\(", "\\\\(");
    regex = regex.replaceAll("\\)", "\\\\)");
    regex = regex.replaceAll("\\{", "\\\\{");
    regex = regex.replaceAll("\\}", "\\\\}");

    return regex;
  }

  private void replace(String search, String replace, boolean global) {

    if (global) {
      filecontents = filecontents.replaceAll(search, replace);
    }
    else {
      filecontents = filecontents.replaceFirst(search, replace);
    }
  }

  private String readfile(File file) {

    StringBuffer contents = new StringBuffer();

    try {
      BufferedReader in = new BufferedReader(new FileReader(file));
      String readline;
      while ((readline = in.readLine()) != null) {
        contents.append(readline);
        contents.append("\n");
      }
      in.close();
    }
    catch (IOException e) {
      System.out.println("Error reading " + file.getName());
      System.exit(0);
    }
    return contents.toString();
  }

  private Document readXML(String XMLstring) {

    Document doc = null;
    InputSource XMLinput = new InputSource(new StringReader(XMLstring));

    try {
      DocumentBuilderFactory docfactory = DocumentBuilderFactory.newInstance();
      docfactory.setValidating(false);
      doc = docfactory.newDocumentBuilder().parse(XMLinput);
    }
    catch (Exception e) {
      System.out.println(e);
    }
    return doc;
  }

  private void printhelp() {

    String hlp = "Usage: PCGentoWerks [OPTIONS] [FILE]\n";
    hlp += "Process character sheet FILE for import in kLoOge.Werks 2.0.\n\n";
    hlp += "FILE has to be created by PCGen with the kLoOge.Werks template.\n";
    hlp += "FILE will be overwritten unless you specify an output directory.\n\n";
    hlp += "Options:\n";
    hlp += "  -o [OUTPUTDIR]     Write to directory OUTPUTDIR instead of overwriting FILE.\n";
    hlp += "  -d                 Delete FILE after processing (OUTPUTDIR must be set).\n";
    hlp += "  -p                 Insert placeholders where possible.\n";
    hlp += "                     See readme.txt for details.\n";
    hlp += "  -f                 Force processing of FILE even if it isn't created with\n";
    hlp += "                     the kLoOge.Werks template.\n";
    hlp += "  -s [WERKSDIR]      Add spell sounds from .wav files in [WERKSDIR]/sounds.\n";
    hlp += "  -nobooks           Don't output spellbooks.\n";
    hlp += "  -nosaves           Don't output spell saving throw information (see readme).\n";
    hlp += "  -nospelllevelnames Don't add class names to spell levels (see readme).\n";
    hlp += "  -v                 Print version info.\n";

    System.out.print(hlp);
  }

  private void printversion() {

    System.out.println("Version 0.8.1, Feb 2005.");
    System.out.println("Made by Cador <cador@tabularum.dk>");

  }
}